"""Utility functionality for tabor awg interaction.

    Copyright (C) 2021 Simon Humpohl

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

    Partially derived from teawg by 2016 Tabor-Electronics Ltd. distributed under GPL 2
    http://www.taborelec.com/
"""

import socket
import logging
import dataclasses
from packaging import version

from typing import Iterator, Optional, Mapping, Tuple

BROADCAST = "255.255.255.255"
IDN_TEMPLATE = "{manuf_name},{model_name},{serial_nb},{fw_ver}"

module_logger = logging.getLogger(__name__)


@dataclasses.dataclass
class DeviceProperties:
    model_name: str
    fw_ver: version.Version
    serial_num: str

    num_parts: int        # number of instrument parts
    chan_per_part: int    # number of channels per part
    seg_quantum: int       # segment-length quantum
    min_seg_len: int      # minimal segment length
    max_arb_mem: int      # maximal arbitrary-memory (points per channel)
    min_dac_val: int         # minimal DAC value
    max_dac_val: int   # maximal DAC value
    max_num_segs: int     # maximal number of segments
    max_seq_len: int   # maximal sequencer-table length (# rows)
    min_seq_len: int         # minimal sequencer-table length (# rows)
    max_num_seq: int      # maximal number of sequencer-table
    max_aseq_len: int # maximal advanced-sequencer table length
    min_aseq_len: int  # minimal advanced-sequencer table length
    min_sclk: float      # minimal sampling-rate (samples/seconds)
    max_sclk: float   # maximal sampling-rate (samples/seconds)
    digital_support: bool   # is digital-wave supported?

    def as_dict(self):
        return dataclasses.asdict(self)


# WX2184 Properties
_wx2184_properties = {
    'model_name'      : 'WX2184', # the model name
    'fw_ver'          : 0.0,      # the firmware version
    'serial_num'      : '0'*9,    # serial number
    'num_parts'       : 2,        # number of instrument parts
    'chan_per_part'   : 2,        # number of channels per part
    'seg_quantum'     : 16,       # segment-length quantum
    'min_seg_len'     : 192,      # minimal segment length
    'max_arb_mem'     : 32 * 10**6,     # maximal arbitrary-memory (points per channel)
    'min_dac_val'     : 0,        # minimal DAC value
    'max_dac_val'     : 2**14-1,  # maximal DAC value
    'max_num_segs'    : 32E+3,    # maximal number of segments
    'max_seq_len'     : 48*1024,  # maximal sequencer-table length (# rows)
    'min_seq_len'     : 3,        # minimal sequencer-table length (# rows)
    'max_num_seq'     : 1000,     # maximal number of sequencer-table
    'max_aseq_len'    : 48*1024-2,# maximal advanced-sequencer table length
    'min_aseq_len'    : 3,        # minimal advanced-sequencer table length
    'min_sclk'        : 75e6,     # minimal sampling-rate (samples/seconds)
    'max_sclk'        : 2300e6,   # maximal sampling-rate (samples/seconds)
    'digital_support' : False,    # is digital-wave supported?
    }

# WX1284 Definitions
_wx1284_properties = {
    'model_name'      : 'WX1284', # the model name
    'fw_ver'          : 0.0,      # the firmware version
    'serial_num'      : '0'*9,    # serial number
    'num_parts'       : 2,        # number of instrument parts
    'chan_per_part'   : 2,        # number of channels per part
    'seg_quantum'     : 16,       # segment-length quantum
    'min_seg_len'     : 192,      # minimal segment length
    'max_arb_mem'     : 32 * 10**6,     # maximal arbitrary-memory (points per channel)
    'min_dac_val'     : 0,        # minimal DAC value
    'max_dac_val'     : 2**14-1,  # maximal DAC value
    'max_num_segs'    : 32 * 10**3,    # maximal number of segments
    'max_seq_len'     : 48*1024,  # maximal sequencer-table length (# rows)
    'min_seq_len'     : 3,        # minimal sequencer-table length (# rows)
    'max_num_seq'     : 1000,     # maximal number of sequencer-table
    'max_aseq_len'    : 48*1024-2,# maximal advanced-sequencer table length
    'min_aseq_len'    : 3,        # minimal advanced-sequencer table length
    'min_sclk'        : 75e6,     # minimal sampling-rate (samples/seconds)
    'max_sclk'        : 1250e6,   # maximal sampling-rate (samples/seconds)
    'digital_support' : False,    # is digital-wave supported?
    }

# WX2182C Definitions
_wx2182C_properties = {
    'model_name'      : 'WX2182C',# the model name
    'fw_ver'          : 0.0,      # the firmware version
    'serial_num'      : '0'*9,    # serial number
    'num_parts'       : 2,        # number of instrument parts
    'chan_per_part'   : 1,        # number of channels per part
    'seg_quantum'     : 16,       # segment-length quantum
    'min_seg_len'     : 192,      # minimal segment length
    'max_arb_mem'     : 32 * 10**6,     # maximal arbitrary-memory (points per channel)
    'min_dac_val'     : 0,        # minimal DAC value
    'max_dac_val'     : 2**14-1,  # maximal DAC value
    'max_num_segs'    : 32 * 10**3,    # maximal number of segments
    'max_seq_len'     : 48*1024,  # maximal sequencer-table length (# rows)
    'min_seq_len'     : 3,        # minimal sequencer-table length (# rows)
    'max_num_seq'     : 1000,     # maximal number of sequencer-table
    'max_aseq_len'    : 1000,     # maximal advanced-sequencer table length
    'min_aseq_len'    : 3,        # minimal advanced-sequencer table length
    'min_sclk'        : 10e6,     # minimal sampling-rate (samples/seconds)
    'max_sclk'        : 2.3e9,    # maximal sampling-rate (samples/seconds)
    'digital_support' : False,    # is digital-wave supported?
    }

# WX1282C Definitions
_wx1282C_properties = {
    'model_name'      : 'WX1282C',# the model name
    'fw_ver'          : 0.0,      # the firmware version
    'serial_num'      : '0'*9,    # serial number
    'num_parts'       : 2,        # number of instrument parts
    'chan_per_part'   : 1,        # number of channels per part
    'seg_quantum'     : 16,       # segment-length quantum
    'min_seg_len'     : 192,      # minimal segment length
    'max_arb_mem'     : 32 * 10**6,     # maximal arbitrary-memory (points per channel)
    'min_dac_val'     : 0,        # minimal DAC value
    'max_dac_val'     : 2**14-1,  # maximal DAC value
    'max_num_segs'    : 32 * 10**3,    # maximal number of segments
    'max_seq_len'     : 48*1024,  # maximal sequencer-table length (# rows)
    'min_seq_len'     : 3,        # minimal sequencer-table length (# rows)
    'max_num_seq'     : 1000,     # maximal number of sequencer-table
    'max_aseq_len'    : 1000,     # maximal advanced-sequencer table length
    'min_aseq_len'    : 3,        # minimal advanced-sequencer table length
    'min_sclk'        : 10e6,     # minimal sampling-rate (samples/seconds)
    'max_sclk'        : 1.25e9,   # maximal sampling-rate (samples/seconds)
    'digital_support' : False,    # is digital-wave supported?
    }

# dictionary of supported-models' properties
model_properties_dict = {
    'WX2184'  : _wx2184_properties,
    'WX2184C' : _wx2184_properties,
    'WX1284'  : _wx2184_properties,
    'WX1284C' : _wx2184_properties,
    'WX2182C' : _wx2182C_properties,
    'WX1282C' : _wx1282C_properties,
    }

# Maps SCPI data commands to file-type IDs
# (used by send_load_file_cmd())
_file_types_dict = {
    # Analog Wave Data:
    'TRACE:DATA' : 1,
    'TRAC:DATA'  : 1,
    'TRACE'      : 1,
    'TRAC'       : 1,

    # Digital Wave Data:
    'DIGITAL:DATA' : 2,
    'DIG:DATA'     : 2,

    # Segment Lengths List:
    'SEGMENT:DATA' : 3,
    'SEGM:DATA'    : 3,
    'SEGMENT'      : 3,
    'SEGM'         : 3,

    # Sequencer-Table (of the active sequencer):
    'SEQUENCE:DATA' : 4,
    'SEQ:DATA'      : 4,
    'SEQUENCE'      : 4,
    'SEQ'           : 4,

    # Sequencer-Lines (of all sequencers):
    'SEQUENCE:DATA:ALL' : 5,
    'SEQ:DATA:ALL'      : 5,
    'SEQUENCE:ALL'      : 5,
    'SEQ:ALL'           : 5,

    # Advanced-Sequencer Table:
    'ASEQUENCE:DATA' : 6,
    'ASEQ:DATA'      : 6,
    'ASEQUENCE'      : 6,
    'ASEQ'           : 6,

    # Pattern-Composer Fast (DC) Pattern Table:
    'PATTERN:COMPOSER:FAST:DATA' : 7,
    'PATTERN:COMPOSER:FAST'      : 7,
    'PATTERN:COMP:FAST:DATA'     : 7,
    'PATTERN:COMP:FAST'          : 7,
    'PATT:COMPOSER:FAST:DATA'    : 7,
    'PATT:COMPOSER:FAST'         : 7,
    'PATT:COMP:FAST:DATA'        : 7,
    'PATT:COMP:FAST'             : 7,

    # Pattern-Composer Linear Pattern Table
    'PATTERN:COMPOSER:LINEAR:DATA' : 8,
    'PATTERN:COMPOSER:LINEAR'      : 8,
    'PATTERN:COMPOSER:LIN:DATA'    : 8,
    'PATTERN:COMPOSER:LIN'         : 8,
    'PATTERN:COMP:LINEAR:DATA'     : 8,
    'PATTERN:COMP:LINEAR'          : 8,
    'PATTERN:COMP:LIN:DATA'        : 8,
    'PATTERN:COMP:LIN'             : 8,
    'PATT:COMPOSER:LINEAR:DATA'    : 8,
    'PATT:COMPOSER:LINEAR'         : 8,
    'PATT:COMPOSER:LIN:DATA'       : 8,
    'PATT:COMPOSER:LIN'            : 8,
    'PATT:COMP:LINEAR:DATA'        : 8,
    'PATT:COMP:LINEAR'             : 8,
    'PATT:COMP:LIN:DATA'           : 8,
    'PATT:COMP:LIN'                : 8,

    # Pattern Predefined Data:
    'PATTERN:PREDEFINED:DATA' : 9,
    'PATTERN:PRED:DATA'       : 9,
    'PATTERN:DATA'            : 9,
    'PATT:PREDEFINED:DATA'    : 9,
    'PATT:PRED:DATA'          : 9,
    'PATT:DATA'               : 9,

    # Digital-Pod Parameters
    'DIGITAL:PARAMETERS' : 10,
    'DIGITAL:PAR'        : 10,
    'DIG:PARAMETERS'     : 10,
    'DIG:PAR'            : 10,

    # ASK Table:
    'ASK:DATA' : 11,

    # FSK Table:
    'FSK:DATA' : 12,

    # FHOP Fixed-Time Table:
    'FHOPPING:FIXED:DATA' : 13,
    'FHOPPING:FIX:DATA'   : 13,
    'FHOP:FIXED:DATA'     : 13,
    'FHOP:FIX:DATA'       : 13,

    # FHOP Variable-Time Table:
    'FHOPPING:VARIABLE:DATA' : 14,
    'FHOPPING:VAR:DATA'      : 14,
    'FHOP:VARIABLE:DATA'     : 14,
    'FHOP:VAR:DATA'          : 14,

    # AHOP Fixed-Time Table:
    'AHOPPING:FIXED:DATA' : 15,
    'AHOPPING:FIX:DATA'   : 15,
    'AHOP:FIXED:DATA'     : 15,
    'AHOP:FIX:DATA'       : 15,

    # AHOP Variable-Time Table:
    'AHOPPING:VARIABLE:DATA' : 16,
    'AHOPPING:VAR:DATA'      : 16,
    'AHOP:VARIABLE:DATA'     : 16,
    'AHOP:VAR:DATA'          : 16,

    # PSK Table:
    'PSK:DATA' : 17,

    # PSK User-Defined Phases:
    'PSK:USER:DATA' : 18,

    # (N)QAM Table:
    'QAM:DATA' : 19,

    # (N)QAM User-Defined Constellation Table:
    'QAM:USER:DATA' : 20,

    # Firmware-Update File:
    'SYSTEM:FIRMWARE:DATA' : 21,
    'SYSTEM:FIRM:DATA'     : 21,
    'SYST:FIRMWARE:DATA'   : 21,
    'SYST:FIRM:DATA'       : 21,
    }


def _parse_max_arb_mem_from_opt_str(opt_str: str) -> Optional[int]:
    if opt_str.startswith('1M', 1):
        return 1 * 10**6
    elif opt_str.startswith('2M', 1):
        return 2 * 10**6
    elif opt_str.startswith('8M', 1):
        return 8 * 10**6
    elif opt_str.startswith('16M', 1):
        return 16 * 10**6
    elif opt_str.startswith('32M', 1):
        return 32 * 10**6
    elif opt_str.startswith('64M', 1):
        return 64 * 10**6
    elif opt_str.startswith('512K', 1):
        return 512 * 10**6
    elif opt_str.startswith('116') or opt_str.startswith('216') or opt_str.startswith('416'):
        return 16 * 10**6
    elif opt_str.startswith('132') or opt_str.startswith('232') or opt_str.startswith('432'):
        return 32 * 10**6
    elif opt_str.startswith('164') or opt_str.startswith('264') or opt_str.startswith('464'):
        return 64 * 10**6


def get_device_properties(idn_str: str, opt_str: str) -> DeviceProperties:
    """
    Args:
        idn_str: the instrument's answer to '*IDN?' query.
        opt_str: the instrument's answer to '*OPT?' query.

    Raises:
        ValueError if idn_str or opt_str does not belong to a supported instrument

    Returns:
        device properties
    """

    try:
        manufacturer_name, model_name, serial_number, fw_ver = idn_str.split(',')
    except ValueError as err:
        raise ValueError("Invalid IDN") from err

    try:
        prop_dict = model_properties_dict[model_name].copy()
    except KeyError as err:
        raise ValueError("Unknown model name", model_name) from err

    if model_name in ('WX2184', 'WX2184C', 'WX1284', 'WX1284C'):
        prop_dict['max_arb_mem'] = int(opt_str[2:4]) * 10**6
    else:
        parsed = _parse_max_arb_mem_from_opt_str(opt_str)
        if parsed is not None:
            prop_dict['max_arb_mem'] = parsed

    prop_dict['model_name'] = model_name
    prop_dict['serial_num'] = serial_number
    prop_dict['fw_ver'] = version.parse(fw_ver)

    return DeviceProperties(**prop_dict)


def udp_awg_instruments() -> Iterator[Tuple[str, str]]:
    """Using UDP list all broadcast reachable AWG-Instruments with LAN Interface.

    Returns:
         Mapping: Resource name -> IDN
    """
    UPFRMPORT = 7502
    FRMHEADERLEN = 22
    FRMDATALEN = 1024
    FLASHLINELEN = 32

    logger = module_logger.getChild('udp_awg_instruments')

    query_msg = bytearray([0xff] * FRMHEADERLEN)
    query_msg[0:4] = b'TEID'

    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_IP) as udp_server_sock:
        # udp_server_sock.bind(("0.0.0.0", UDPSRVPORT))  # any IP-Address
        udp_server_sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, FRMHEADERLEN + FRMDATALEN)
        udp_server_sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, True)

        # Send the query-message (to all)
        udp_server_sock.sendto(query_msg, (BROADCAST, UPFRMPORT))

        # Receive responses
        udp_server_sock.settimeout(2)
        while True:
            try:
                raw_data, (ip_addr, port) = udp_server_sock.recvfrom(FRMHEADERLEN + FRMDATALEN)
                resource_name = f"TCPIP::{ip_addr}::5025::SOCKET"
                logger.debug("Received answer from %s:%d: %r", ip_addr, port, raw_data)

                attrs = {}

                data = raw_data[FRMHEADERLEN:]
                while len(data) >= FLASHLINELEN and len(attrs) < 4:
                    line = data[:FLASHLINELEN]
                    data = data[FLASHLINELEN:]

                    op_code = line[:1]
                    attr = line[1:].rstrip(b'\x00')

                    if attr:
                        logger.debug("Attribute of %s with op code %s: %s", ip_addr, op_code, attr)

                    if op_code == b'D':
                        attrs['manuf_name'] = attr.decode('ascii')
                    if op_code == b'I':
                        attrs['model_name'] = attr.decode('ascii')
                    if op_code == b'S':
                        attrs['serial_nb'] = attr.decode('ascii')
                    if op_code == b'F':
                        attrs['fw_ver'] = attr.decode('ascii')

                if len(attrs) == 4:
                    yield resource_name, IDN_TEMPLATE.format(**attrs)
                else:
                    logger.info("Ignoring %s because it send %d != 4 attributes", resource_name, len(attrs))
            except socket.timeout:
                break





"""This module provides an object oriented abstraction.

    Copyright (C) 2021 Simon Humpohl

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

    Partially derived from teawg by 2016 Tabor-Electronics Ltd. distributed under GPL 2
    http://www.taborelec.com/
"""
from typing import Optional, Sequence, Tuple

from packaging import version
import numpy as np

import pyvisa.util
from pyvisa.resources import MessageBasedResource
import pyvisa.constants as vc

import tabor_control
from tabor_control import ParanoiaLevel
from tabor_control.util import get_device_properties, DeviceProperties


class TEWXAwg:
    """Tabor-Electronics WX-Instrument Controller"""

    def __init__(self, visa_instr: MessageBasedResource, paranoia_level: ParanoiaLevel):
        self._visa_inst = visa_instr
        self._dev_props: Optional[DeviceProperties] = None
        self._resource_name = None
        self._paranoia_level = ParanoiaLevel(paranoia_level)
        self._simulator_read_workaround = None

        if visa_instr is not None:
            self._resource_name = visa_instr.resource_name

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def supports_basic_reading(self) -> bool:
        return self.model_name.lower() in ('wx2184', 'wx2184c', 'wx1284', 'wx1284c') and self.fw_ver >= version.Version('3')

    def open(self, resource_name=None):
        self.close()

        if resource_name is not None:
            self._resource_name = resource_name

        if self._resource_name is not None:
            self._visa_inst = tabor_control.open_session(self._resource_name, extra_init=True)

    def close(self):
        """Close Connection and forget saved properties"""
        if self._visa_inst is not None:
            try:
                self._visa_inst.close()
            except pyvisa.errors.Error:
                pass
        self._visa_inst = None
        self._dev_props = None
        self._simulator_read_workaround = None

    @property
    def visa_inst(self) -> MessageBasedResource:
        return self._visa_inst

    @property
    def resource_name(self):
        return self._resource_name

    @property
    def dev_properties(self) -> DeviceProperties:
        '''Get dictionary of the device properties '''
        if self._dev_props is None:
            idn = self._visa_inst.query('*IDN?')
            opt = self._visa_inst.query('*OPT?')

            self._dev_props = get_device_properties(idn, opt)
        return self._dev_props

    @property
    def model_name(self) -> str:
        return self._dev_props.model_name

    @property
    def fw_ver(self) -> version.Version:
        """Firmware version"""
        return self._dev_props.fw_ver

    @property
    def is_simulator(self) -> bool:
        """True if IP address is localhost"""
        if self.visa_inst and self.visa_inst.interface_type == vc.InterfaceType.tcpip:
            return '127.0.0.1' in self.visa_inst.resource_name or 'localhost' in self.visa_inst.resource_name
        return False

    @property
    def paranoia_level(self) -> ParanoiaLevel:
        """Get the (default) paranoia-level"""
        return self._paranoia_level

    @paranoia_level.setter
    def paranoia_level(self, value: ParanoiaLevel):
        """Set the (default) paranoia-level"""
        self._paranoia_level = ParanoiaLevel(value)

    def send_cmd(self, cmd_str: str, paranoia_level=None):
        '''Send the given command to the instrument

        :param cmd_str: the command string (SCPI statement).
        :param paranoia_level: the paranoia-level (overrides the default one)
        '''
        if paranoia_level is None:
            paranoia_level = self._paranoia_level
        else:
            paranoia_level = ParanoiaLevel(paranoia_level)

        if self._simulator_read_workaround is None:
            workaround = self.is_simulator
        else:
            workaround = self._simulator_read_workaround

        tabor_control.send_cmd(self._visa_inst, cmd_str, paranoia_level)
        if workaround and paranoia_level == ParanoiaLevel.NONE:
            empty_str = self._visa_inst.read()
            assert empty_str == '', ("This workaround expects an empty response after each command that is not a query."
                                     " The simulator does this in some versions. It is controlled by the attribute "
                                     "_simulator_read_workaround. Actual response was: %r" % empty_str)

    def send_query(self, query_str) -> str:
        """Send the given query to the instrument and read the response

        :param query_str: the query string (SCPI statement).
        :returns: the instrument's response."""
        return self._visa_inst.query(query_str)

    def write_segment_data(self, waveform_data: np.ndarray, paranoia_level=None):
        if paranoia_level is None:
            paranoia_level = self._paranoia_level
        tabor_control.write_waveform_data(self._visa_inst, waveform_data, paranoia_level)

    def read_segment_data(self) -> np.ndarray:
        """Read data of active segment.

        Raises:
            ReadNotSupported if unsupported

        Returns:

        """
        if self.supports_basic_reading():
            dat = self._visa_inst.query_binary_values(':TRAC:DATA?', datatype='H', container=np.array)
            return np.asarray(dat, dtype=np.uint16)
        else:
            raise ReadNotSupported(self.dev_properties)

    def write_segment_lengths(self, segment_lengths: np.ndarray, paranoia_level: ParanoiaLevel = None):
        if paranoia_level is None:
            paranoia_level = self.paranoia_level
        tabor_control.write_segment_lengths(self._visa_inst,
                                            segment_lengths=segment_lengths,
                                            paranoia_level=paranoia_level)

    def write_sequencer_table(self, sequencer_table: Sequence[Tuple[int, int, int]],
                              paranoia_level: ParanoiaLevel = None):
        if paranoia_level is None:
            paranoia_level = self.paranoia_level
        tabor_control.write_sequencer_table(self._visa_inst, sequencer_table, paranoia_level)

    def read_sequencer_table(self):
        """Read the active sequencer table (of the active part).

        Returns:
            numpy structured array with dtype tabor_control.TABLE_ENTRY_DTYPE
        """
        if not self.supports_basic_reading():
            raise ReadNotSupported(self.dev_properties)
        else:
            return tabor_control.read_sequencer_table(self._visa_inst)

    def write_advanced_sequencer_table(self, advanced_sequencer_table: Sequence[Tuple[int, int, int]],
                                       paranoia_level: ParanoiaLevel = None):
        if paranoia_level is None:
            paranoia_level = self.paranoia_level
        tabor_control.write_advanced_sequencer_table(self._visa_inst, advanced_sequencer_table, paranoia_level)

    def read_advanced_sequencer_table(self) -> np.ndarray:
        """Read the advanced sequencer table (of the active part).

        Raises:
            ReadNotSupported if the device cannot be read from
        Returns:
            numpy structured array with dtype tabor_control.TABLE_ENTRY_DTYPE
        """
        if not self.supports_basic_reading():
            raise ReadNotSupported(self.dev_properties)
        else:
            return tabor_control.read_advanced_sequencer_table(self._visa_inst)


class ReadNotSupported(RuntimeError):
    """Raised if a read is attempted on a device that does not support it"""



"""
    tabor_control - Interact with arbitrary waveform generators from Tabor-Electronics Ltd.
    Copyright (C) 2021 Simon Humpohl

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

    Partially derived from pyte16 by 2016 Tabor-Electronics Ltd. distributed under GPL 2
    http://www.taborelec.com/
"""

from typing import Sequence, Tuple, cast
import socket
import enum
import warnings
import struct

import numpy as np

from pyvisa import ResourceManager
from pyvisa.resources import Resource, MessageBasedResource
import pyvisa.constants as vc


__version__ = "0.1"


TABLE_ENTRY_FORMAT = struct.Struct('< L H B x')
TABLE_ENTRY_DTYPE = np.dtype([('repeats', 'L'), ('segment_no', 'H'), ('jump_flag', 'B')], align=True)
assert TABLE_ENTRY_FORMAT.size == TABLE_ENTRY_DTYPE.itemsize


class ParanoiaLevel(enum.Enum):
    NONE = 0
    OPC = 1
    SYST_ERR = 2


def init_vi_inst(vi: Resource,
                 timeout_msec: int = 30000,
                 read_buff_size_bytes: int = 4096,
                 write_buff_size_bytes: int = 4096):
    """Initialize the given Instrument VISA Session"""

    vi.timeout = int(timeout_msec)
    vi.visalib.set_buffer(vi.session, vc.VI_READ_BUF, int(read_buff_size_bytes))
    vi.visalib.set_buffer(vi.session, vc.VI_WRITE_BUF, int(write_buff_size_bytes))
    vi.read_termination = '\n'
    vi.write_termination = '\n'
    intf_type = vi.get_visa_attribute(vc.VI_ATTR_INTF_TYPE)
    if intf_type in (vc.VI_INTF_USB, vc.VI_INTF_GPIB, vc.VI_INTF_TCPIP, vc.VI_INTF_ASRL):
        vi.set_visa_attribute(vc.VI_ATTR_WR_BUF_OPER_MODE, vc.VI_FLUSH_ON_ACCESS)
        vi.set_visa_attribute(vc.VI_ATTR_RD_BUF_OPER_MODE, vc.VI_FLUSH_ON_ACCESS)
        if intf_type == vc.VI_INTF_TCPIP:
            vi.set_visa_attribute(vc.VI_ATTR_TERMCHAR_EN, vc.VI_TRUE)  # vc.VI_FALSE
        elif intf_type == vc.VI_INTF_ASRL:
            vi.set_visa_attribute(vc.VI_ATTR_ASRL_BAUD, 115200)
            vi.set_visa_attribute(vc.VI_ATTR_ASRL_END_OUT, 0)
            vi.set_visa_attribute(vc.VI_ATTR_ASRL_END_IN, 2)
    vi.clear()


def _validate_syst_err_answer(query_str, syst_err: str):
    try:
        err_num, err_txt = syst_err.split(',')
        err_num = int(err_num)
    except (ValueError, TypeError):
        warnings.warn(IllformedAnswer(query_str, syst_err))
    else:
        if err_num != 0:
            warnings.warn(SystErr(query_str, err_num, err_txt))


def open_session(resource_name: str, resource_manager: ResourceManager = None, extra_init=True) -> MessageBasedResource:
    """Open VISA Session (optionally prompt for resource name).

    The `resource_name` can be either:
        1. Full VISA Resource-Name (e.g. 'TCPIP::192.168.0.170::5025::SOCKET')
        2. IP-Address (e.g. '192.168.0.170')
    """

    if resource_manager is None:
        resource_manager = ResourceManager()

    try:
        packed_ip = socket.inet_aton(resource_name)
    except OSError:
        pass
    else:
        resource_name = f"TCPIP::{socket.inet_ntoa(packed_ip)}::5025::SOCKET"

    vi = resource_manager.open_resource(resource_name)
    if extra_init:
        init_vi_inst(vi)

    return vi


def send_cmd(vi: MessageBasedResource, cmd_str: str, paranoia_level: ParanoiaLevel = ParanoiaLevel.OPC):
    """

    Args:
        vi:
        cmd_str:
        paranoia_level:

    Raises:
        ValueError: If '?' is in cmd_str

    Returns:

    """
    paranoia_level = ParanoiaLevel(paranoia_level)
    if '?' in cmd_str:
        raise ValueError("? in command string. Use send_query.")
    cmd_str = cmd_str.rstrip()
    cmd_list = [cmd_str] if cmd_str else []

    if paranoia_level == ParanoiaLevel.NONE:
        vi.write(cmd_str)

    elif paranoia_level == ParanoiaLevel.OPC:
        cmd_list.append('*OPC?')
        query = ';'.join(cmd_list)
        opc = vi.query(query)

        if opc != '1':
            warnings.warn(OPCWarning(query, opc))

    elif paranoia_level == ParanoiaLevel.SYST_ERR:
        cmd_list.append(':SYST:ERR?')
        query = ';'.join(cmd_list)
        syst_err = vi.query(query)
        _validate_syst_err_answer(query, syst_err)

    else:
        raise ValueError("Unhandled paranoia level")


def _pre_download_binary_data(vi, bin_dat_size=None):
    '''Pre-Download Binary-Data

    :param vi: `pyvisa` instrument.
    :param bin_dat_size: the binary-data-size in bytes (can be omitted)
    :returns: the max write-chunk size (in bytes) and the original time-out (in msec)
    '''
    orig_timeout = vi.timeout

    max_chunk_size = vi.write_buff_size if hasattr(vi, 'write_buff_size') else 4096

    try:
        intf_type = vi.get_visa_attribute(vc.VI_ATTR_INTF_TYPE)
        if intf_type == vc.VI_INTF_GPIB:
            _ = vi.write("*OPC?")
            for _ in range(2000):
                status_byte = vi.stb
                if (status_byte & 0x10) == 0x10:
                    break
            _ = vi.read()
            max_chunk_size = min(max_chunk_size, 30000)
            if bin_dat_size is not None and orig_timeout < bin_dat_size / 20:
                vi.timeout = int(bin_dat_size / 20)
        else:
            max_chunk_size = min(max_chunk_size, 256000)
    except:
        pass

    return orig_timeout, max_chunk_size


def _write_binary_data(vi: MessageBasedResource, prefix: str, data: bytes, paranoia_level: ParanoiaLevel):
    paranoia_level = ParanoiaLevel(paranoia_level)

    assert isinstance(data, bytes)

    data_len = f"{len(data)}"
    header = f'{prefix}#{len(data_len)}{data_len}'.encode()

    written = vi.write_raw(header)
    if written != len(header):
        raise RuntimeError("Binary header not completely written. Probably unrecoverable error.")

    # TODO: do we need manual chunked writing here?
    written = vi.write_raw(data)
    if written != len(data):
        raise RuntimeError("Binary data not completely written. Probably unrecoverable error.")

    if paranoia_level == ParanoiaLevel.NONE:
        pass

    elif paranoia_level == ParanoiaLevel.OPC:
        opc = vi.query('*OPC?')
        if opc != '1':
            warnings.warn(OPCWarning('*OPC?', opc))

    else:
        query = ':SYST:ERR?'
        syst_err = vi.query(query)
        _validate_syst_err_answer(query, syst_err)


def write_waveform_data(vi: MessageBasedResource, data: np.ndarray, paranoia_level: ParanoiaLevel = ParanoiaLevel.OPC):
    if not (isinstance(data, np.ndarray) and data.dtype.char != np.uint16):
        raise TypeError("Must be a numpy array of unsigned 16 bit integers")

    if not (data.ndim == 1 and data.size >= 192 and data.size % 16 == 0):
        raise ValueError("Data must have the correct dimensions")

    b_data = data.tobytes()
    _write_binary_data(vi, ':TRAC:DATA', b_data, paranoia_level)


def write_segment_lengths(vi, segment_lengths: np.ndarray, paranoia_level: ParanoiaLevel = ParanoiaLevel.OPC):
    if not (isinstance(segment_lengths, np.ndarray) and segment_lengths.dtype == np.uint32):
        raise TypeError("Segment lengths must be a numpy array of unsigned 32 bit integers")

    if not (segment_lengths.ndim == 1):
        raise ValueError("Segment lengths must be a one dimensional array.")

    b_data = segment_lengths.tobytes()
    _write_binary_data(vi, ':SEGM:DATA', b_data, paranoia_level)


def _write_table(vi: MessageBasedResource, prefix: str, table: Sequence[Tuple[int, int, int]],
                 paranoia_level: ParanoiaLevel):
    table_entry_format = TABLE_ENTRY_FORMAT
    entry_size = table_entry_format.size

    table_len = len(table)
    b_data = bytearray(table_len * entry_size)

    for pos, (repeats, seg_nb, jump_flag) in enumerate(table):
        table_entry_format.pack_into(b_data, pos * entry_size, repeats, seg_nb, jump_flag)

    _write_binary_data(vi, prefix, data=bytes(b_data), paranoia_level=paranoia_level)


def _read_table(vi: MessageBasedResource, query_cmd: str) -> np.ndarray:
    data_raw = cast(np.ndarray, vi.query_binary_values(query_cmd, datatype='B', container=np.array))
    return data_raw.view(TABLE_ENTRY_DTYPE)


def write_sequencer_table(vi, sequencer_table: Sequence[Tuple[int, int, int]],
                          paranoia_level: ParanoiaLevel = ParanoiaLevel.OPC):
    """Write Sequencer-Table to Instrument

    The sequencer-table, `seq_table`, is a list of 3-tuples
    of the form: (<repeats>, <segment no.>, <jump-flag>)

    Example:
        >>> # Create Sequencer-Table:
        >>> repeats = [ 1, 1, 100, 4, 1 ]
        >>> seg_nb = [ 2, 3, 5, 1, 4 ]
        >>> jump = [ 0, 0, 1, 0, 0 ]
        >>> my_sequencer_table = zip(repeats, seg_nb, jump)
        >>>
        >>> # Select sequence no. 1:
        >>>  _ = vi.ask(':SEQ:SELect 1; *OPC?')
        >>>
        >>> # Download the sequencer-table:
        >>> tabor_control.write_sequencer_table(vi, sequencer_table)
    """
    _write_table(vi, ':SEQ:DATA', sequencer_table, paranoia_level=paranoia_level)


def read_sequencer_table(vi: MessageBasedResource) -> np.ndarray:
    return _read_table(vi, ':SEQ:DATA?')


def write_advanced_sequencer_table(vi, advanced_sequencer_table: Sequence[Tuple[int, int, int]],
                                   paranoia_level: ParanoiaLevel = ParanoiaLevel.OPC):
    """Download Advanced-Sequencer-Table to Instrument

    The advanced-sequencer-table, `adv_seq_table`, is a list of 3-tuples
    of the form: (<repeats>, <sequence no.>, <jump-flag>)
    """

    _write_table(vi, ':ASEQ:DATA', table=advanced_sequencer_table, paranoia_level=paranoia_level)


def read_advanced_sequencer_table(vi: MessageBasedResource) -> np.ndarray:
    return _read_table(vi, ':ASEQ:DATA?')


class TaborWarning(RuntimeWarning):
    pass


class OPCWarning(TaborWarning):
    """Warning emitted if answer to OPC is not '1'"""

    def __init__(self, query_str, answer):
        super().__init__(query_str, answer)

    @property
    def query_str(self):
        return self.args[0]

    @property
    def answer(self):
        return self.args[1]

    def __str__(self):
        return f"Received {self.answer!r} as an answer on {self.query_str!r} instead of '1'"


class SystErr(TaborWarning):
    """Emitted if :SYST:ERR? returns a non 0 status code"""
    def __init__(self, query_str, status_code, error_description):
        super().__init__(query_str, status_code, error_description)

    @property
    def query_str(self):
        return self.args[0]

    @property
    def status_code(self) -> int:
        return self.args[1]

    @property
    def error_description(self) -> int:
        return self.args[2]

    def __str__(self):
        return f"Received {self.status_code}, {self.error_description} as an answer on {self.query_str!r} instead of '0, \"No Error\"'"


class IllformedAnswer(TaborWarning):
    """Received an answer in an unexpected format."""
    def __init__(self, query_str, answer):
        super().__init__(query_str, answer)


warnings.simplefilter("error", OPCWarning)
warnings.simplefilter("error", IllformedAnswer)
